;; Generated package description from bnf-mode.el  -*- no-byte-compile: t -*-
(define-package "bnf-mode" "0.4.5.0.20240915.211838" "Major mode for editing BNF grammars." '((cl-lib "0.5") (emacs "27.1")) :commit "5304ab647e04916c5be4fdde41477ad429a89120" :authors '(("Serghei Iakovlev" . "gnu@serghei.pl")) :maintainer '("Serghei Iakovlev" . "gnu@serghei.pl") :keywords '("languages") :url "https://github.com/sergeyklay/bnf-mode")
